package ch.vanwa.lox_interpreter;

import java.util.HashMap;
import java.util.Map;

/**
 * Environment mapping variable names to their values.
 */
class Environment {
  /**
   * Enclosing parent environment, can be null (only used for the outermost global
   * environment).
   */
  final Environment enclosing;

  /**
   * Mapping of variable names to their values.
   */
  private final Map<String, Object> values = new HashMap<>();

  /**
   * Initialise environment without an enclosing environment.
   */
  Environment() {
    enclosing = null;
  }

  /**
   * Initialise environment with an enclosing environment.
   */
  Environment(Environment enclosing) {
    this.enclosing = enclosing;
  }

  /**
   * Get the value of a variable. Throw an error if that variable is not known in
   * this environment or in any of the enclosing ones.
   *
   * @param name Token with the variable lexeme used as variable name.
   */
  Object get(Token name) {
    if (values.containsKey(name.lexeme())) {
      return values.get(name.lexeme());
    }

    if (enclosing != null) {
      return enclosing.get(name);
    }

    throw new RuntimeError(name,
        "Undefined variable '" + name.lexeme() + "'.");
  }

  /**
   * Assign a value to a variable. Throw an error if that variable is not known in
   * this environment or in any of the enclosing ones.
   */
  void assign(Token name, Object value) {
    if (values.containsKey(name.lexeme())) {
      values.put(name.lexeme(), value);
      return;
    }

    if (enclosing != null) {
      enclosing.assign(name, value);
      return;
    }

    throw new RuntimeError(name,
        "Undefined variable '" + name.lexeme() + "'.");
  }

  /**
   * Define a new variable in this environment.
   */
  void define(String name, Object value) {
    values.put(name, value);
  }

  /**
   * Get a variable value at the n-th parent environment of the current one.
   */
  Object getAt(int distance, String name) {
    return ancestor(distance).values.get(name);
  }

  /**
   * Assign a variable value at the n-th parent environment of the current one.
   */
  void assignAt(int distance, Token name, Object value) {
    ancestor(distance).values.put(name.lexeme(), value);
  }

  /**
   * Get the n-th parent environment of the current one.
   */
  Environment ancestor(int distance) {
    Environment environment = this;
    for (int i = 0; i < distance; i++) {
      environment = environment.enclosing;
    }

    return environment;
  }
}
