package ch.vanwa.lox_interpreter;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static ch.vanwa.lox_interpreter.TokenType.*;

/**
 * Convert source code into tokens for the Lox programming language.
 */
class Scanner {
  /**
   * Source code getting scanned.
   */
  private final String source;

  /**
   * List of already scanned tokens.
   */
  private final List<Token> tokens = new ArrayList<>();

  /**
   * Starting index of the current lexeme.
   */
  private int start = 0;

  /**
   * Index of current character in the source code.
   */
  private int current = 0;

  /**
   * Which line the scanneris on in the source code.
   */
  private int line = 1;

  /**
   * Mapping of keyword string to {@link TokenType}.
   */
  private static final Map<String, TokenType> keywords = Map.ofEntries(
      Map.entry("and", AND),
      Map.entry("class", CLASS),
      Map.entry("else", ELSE),
      Map.entry("false", FALSE),
      Map.entry("for", FOR),
      Map.entry("fun", FUN),
      Map.entry("if", IF),
      Map.entry("nil", NIL),
      Map.entry("or", OR),
      Map.entry("print", PRINT),
      Map.entry("return", RETURN),
      Map.entry("super", SUPER),
      Map.entry("this", THIS),
      Map.entry("true", TRUE),
      Map.entry("var", VAR),
      Map.entry("while", WHILE));

  /**
   * Initialize scanner.
   *
   * @param source Source code to scan.
   */
  Scanner(final String source) {
    this.source = source;
  }

  /**
   * Scan lexemes to tokens until the end is reached.
   *
   * @return List of scanned tokens + EOF.
   */
  List<Token> scanTokens() {
    while (!isAtEnd()) {
      start = current;
      scanToken();
    }

    tokens.add(new Token(EOF, "", null, line));
    return tokens;
  }

  /**
   * Scan the next lexeme to a token.
   */
  private void scanToken() {
    char c = advance();
    switch (c) {
      case '(':
        addToken(LEFT_PAREN);
        break;
      case ')':
        addToken(RIGHT_PAREN);
        break;
      case '{':
        addToken(LEFT_BRACE);
        break;
      case '}':
        addToken(RIGHT_BRACE);
        break;
      case ',':
        addToken(COMMA);
        break;
      case '.':
        addToken(DOT);
        break;
      case '-':
        addToken(MINUS);
        break;
      case '+':
        addToken(PLUS);
        break;
      case ';':
        addToken(SEMICOLON);
        break;
      case '*':
        addToken(STAR);
        break;
      case '!':
        addToken(match('=') ? BANG_EQUAL : BANG);
        break;
      case '=':
        addToken(match('=') ? EQUAL_EQUAL : EQUAL);
        break;
      case '<':
        addToken(match('=') ? LESS_EQUAL : LESS);
        break;
      case '>':
        addToken(match('=') ? GREATER_EQUAL : GREATER);
        break;
      case '/':
        if (match('/')) {
          // A comment goes until the end of the line.
          while (peek() != '\n' && !isAtEnd())
            advance();
        } else {
          addToken(SLASH);
        }
        break;
      case ' ':
      case '\r':
      case '\t':
        // Ignore whitespace.
        break;
      case '\n':
        line++;
        break;
      case '"':
        string();
        break;
      default:
        if (isDigit(c)) {
          number();
        } else if (isAlpha(c)) {
          identifier();
        } else {
          Lox.error(line, "Unexpected character.");
        }
        break;
    }
  }

  /**
   * Consume an identifier.
   */
  private void identifier() {
    while (isAlphaNumeric(peek())) {
      advance();
    }

    String text = source.substring(start, current);
    TokenType type = keywords.get(text);
    if (type == null) {
      type = IDENTIFIER;
    }
    addToken(type);
  }

  /**
   * Consume a number literal.
   */
  private void number() {
    while (isDigit(peek())) {
      advance();
    }

    // Look for a fractional part.
    if (peek() == '.' && isDigit(peekNext())) {
      // Consume the "."
      advance();

      while (isDigit(peek())) {
        advance();
      }
    }

    addToken(NUMBER,
        Double.parseDouble(source.substring(start, current)));
  }

  /**
   * Consume a string literal.
   */
  private void string() {
    while (peek() != '"' && !isAtEnd()) {
      if (peek() == '\n') {
        line++;
      }
      advance();
    }

    if (isAtEnd()) {
      Lox.error(line, "Unterminated string.");
      return;
    }

    // The closing ".
    advance();

    // Trim the surrounding quotes.
    String value = source.substring(start + 1, current - 1);
    addToken(STRING, value);
  }

  /**
   * Check if a character matches the one at the current pointer.
   *
   * @param expected Character used for matching against the current character.
   *
   * @return True if the expected character matches the current one, false
   *         otherwise.
   */
  private boolean match(final char expected) {
    if (isAtEnd()) {
      return false;
    }
    if (source.charAt(current) != expected) {
      return false;
    }

    current++;
    return true;
  }

  /**
   * Peek at the current character without consuming it.
   *
   * @return The character peeked at.
   */
  private char peek() {
    if (isAtEnd()) {
      return '\0';
    }
    return source.charAt(current);
  }

  /**
   * Peek at the next character without consuming it or moving the pointer.
   *
   * @return The character peeked at.
   */
  private char peekNext() {
    if (current + 1 >= source.length()) {
      return '\0';
    }
    return source.charAt(current + 1);
  }

  /**
   * Check if a character is a letter or an underscore (a-z,A-Z,_).
   *
   * @param c Character to check.
   *
   * @return True if the character is a letter or underscore, false otherwise.
   */
  private boolean isAlpha(final char c) {
    return (c >= 'a' && c <= 'z') ||
        (c >= 'A' && c <= 'Z') ||
        c == '_';
  }

  /**
   * Check if a character is alphanumeric (a-z,A-Z,_,0-9).
   *
   * @param c Character to check.
   *
   * @return True if the character is alphanumeric, false otherwise.
   */
  private boolean isAlphaNumeric(final char c) {
    return isAlpha(c) || isDigit(c);
  }

  /**
   * Check if a character is a digit (0-9).
   *
   * @param c Character to check.
   *
   * @return True if the character is a digit, false otherwise.
   */
  private boolean isDigit(final char c) {
    return c >= '0' && c <= '9';
  }

  /**
   * Check if the scanner reached the end of the source code.
   *
   * @return True if the scaner reached the end of the source code, false
   *         otherwise.
   */
  private boolean isAtEnd() {
    return current >= source.length();
  }

  /**
   * Advance the source code pointer by one.
   *
   * @return The character under the advanced pointer.
   */
  private char advance() {
    return source.charAt(current++);
  }

  /**
   * Add a token without a literal to the list of scanned tokens.
   *
   * @param type Type of token.
   */
  private void addToken(final TokenType type) {
    addToken(type, null);
  }

  /**
   * Add a token to the list of scanned tokens.
   *
   * @param type    Type of token.
   * @param literal Value of a literal (can be null).
   */
  private void addToken(final TokenType type, final Object literal) {
    String text = source.substring(start, current);
    tokens.add(new Token(type, text, literal, line));
  }
}
