use crate::{expression::Expression, interpreter::Interpreter, statement::Statement, token::Token};
use std::collections::HashMap;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum ResolverError {
    #[error("line {0}: can not read local variable in its own initializer.")]
    ReadVarInOwnInitializer(usize),
    #[error("line {0}: already a variable with this name in this scope.")]
    NameRedefinition(usize),
    #[error("line {0}: can not return from top-level code.")]
    TopLevelReturn(usize),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum FunctionType {
    None,
    Function,
}

#[derive(Debug)]
pub struct Resolver<'a> {
    scopes: Vec<HashMap<String, bool>>,
    interpreter: &'a mut Interpreter,
    current_fun: FunctionType,
}

impl<'a> Resolver<'a> {
    pub fn new(interpreter: &'a mut Interpreter) -> Self {
        Self {
            scopes: Vec::default(),
            interpreter,
            current_fun: FunctionType::None,
        }
    }

    pub fn resolve(&mut self, statements: &[Statement]) -> Result<(), ResolverError> {
        for stm in statements {
            self.resolve_statement(stm)?;
        }
        Ok(())
    }

    fn resolve_statement(&mut self, statement: &Statement) -> Result<(), ResolverError> {
        match statement {
            Statement::Block(statements) => self.resolve_block(statements),
            Statement::Print(expression) => self.resolve_expression(expression),
            Statement::Expression(expression) => self.resolve_expression(expression),
            Statement::Var { name, initializer } => self.resolve_var(name, *initializer.clone()),
            Statement::If {
                condition,
                then_branch,
                else_branch,
            } => self.resolve_if_stmt(
                condition,
                then_branch,
                else_branch.as_ref().map(|boxed| *boxed.clone()),
            ),
            Statement::While { condition, body } => self.resolve_while_stmt(condition, body),
            Statement::Function { name, params, body } => {
                self.resolve_function_stmt(name, params, body)
            }
            Statement::Return { keyword, value } => {
                if self.current_fun == FunctionType::None {
                    return Err(ResolverError::TopLevelReturn(keyword.line));
                }
                if let Some(value) = value {
                    self.resolve_expression(value)
                } else {
                    Ok(())
                }
            }
        }
    }

    fn resolve_expression(&mut self, expr: &Expression) -> Result<(), ResolverError> {
        match expr {
            Expression::Assign { name, value } => self.resolve_assign_expr(expr, name, value),
            Expression::Binary {
                left,
                operator: _,
                right,
            } => self.resolve_binary_expr(left, right),
            Expression::Call {
                callee,
                paren: _,
                args,
            } => self.resolve_call_expr(callee, args),
            Expression::Grouping { expression } => self.resolve_expression(expression),
            Expression::Literal { value: _ } => Ok(()),
            Expression::Logical {
                left,
                operator: _,
                right,
            } => self.resolve_binary_expr(left, right),
            Expression::Unary { operator: _, right } => self.resolve_expression(right),
            Expression::Variable { name } => self.resolve_var_expr(expr, name),
        }
    }

    fn begin_scope(&mut self) {
        self.scopes.push(HashMap::new());
    }

    fn end_scope(&mut self) {
        self.scopes.pop();
    }

    fn declare(&mut self, name: &Token) -> Result<(), ResolverError> {
        if let Some(scope) = self.scopes.last_mut() {
            if scope.contains_key(&name.lexeme) {
                return Err(ResolverError::NameRedefinition(name.line));
            }

            scope.insert(name.lexeme.clone(), false);
        }

        Ok(())
    }

    fn define(&mut self, name: &Token) {
        if let Some(scope) = self.scopes.last_mut() {
            scope.insert(name.lexeme.clone(), true);
        }
    }

    fn resolve_block(&mut self, statements: &[Statement]) -> Result<(), ResolverError> {
        self.begin_scope();
        self.resolve(statements)?;
        self.end_scope();

        Ok(())
    }

    fn resolve_var(
        &mut self,
        name: &Token,
        initializer: Option<Expression>,
    ) -> Result<(), ResolverError> {
        self.declare(name)?;

        if let Some(initializer) = initializer {
            self.resolve_expression(&initializer)?;
        }

        self.define(name);
        Ok(())
    }

    fn resolve_var_expr(&mut self, expr: &Expression, name: &Token) -> Result<(), ResolverError> {
        if let Some(scope) = self.scopes.last() {
            if !scope.get(&name.lexeme).unwrap_or(&true) {
                return Err(ResolverError::ReadVarInOwnInitializer(name.line));
            }
        }
        self.resolve_local(expr, name)?;
        Ok(())
    }

    fn resolve_assign_expr(
        &mut self,
        expr: &Expression,
        name: &Token,
        val: &Expression,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(val)?;
        self.resolve_local(expr, name)?;
        Ok(())
    }

    fn resolve_binary_expr(
        &mut self,
        left: &Expression,
        right: &Expression,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(left)?;
        self.resolve_expression(right)?;
        Ok(())
    }

    fn resolve_function_stmt(
        &mut self,
        name: &Token,
        params: &[Token],
        body: &[Statement],
    ) -> Result<(), ResolverError> {
        self.declare(name)?;
        self.define(name);
        self.resolve_function(params, body, FunctionType::Function)?;

        Ok(())
    }

    fn resolve_call_expr(
        &mut self,
        callee: &Expression,
        args: &[Expression],
    ) -> Result<(), ResolverError> {
        self.resolve_expression(callee)?;

        for arg in args {
            self.resolve_expression(arg)?;
        }

        Ok(())
    }

    fn resolve_if_stmt(
        &mut self,
        condition: &Expression,
        then_branch: &Statement,
        else_branch: Option<Statement>,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(condition)?;
        self.resolve_statement(then_branch)?;

        if let Some(stmt) = else_branch {
            self.resolve_statement(&stmt)?;
        }

        Ok(())
    }

    fn resolve_while_stmt(
        &mut self,
        condition: &Expression,
        body: &Statement,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(condition)?;
        self.resolve_statement(body)?;

        Ok(())
    }

    fn resolve_function(
        &mut self,
        params: &[Token],
        body: &[Statement],
        fun_type: FunctionType,
    ) -> Result<(), ResolverError> {
        let enclosing_fun = self.current_fun;
        self.current_fun = fun_type;

        self.begin_scope();

        for param in params {
            self.declare(param)?;
            self.define(param);
        }
        self.resolve(body)?;

        self.end_scope();

        self.current_fun = enclosing_fun;

        Ok(())
    }

    fn resolve_local(
        &mut self,
        expression: &Expression,
        name: &Token,
    ) -> Result<(), ResolverError> {
        for (idx, scope) in self.scopes.iter().rev().enumerate() {
            if scope.contains_key(&name.lexeme) {
                self.interpreter.resolve(expression.clone(), idx);
                return Ok(());
            }
        }

        Ok(())
    }
}
